import torch
import torchvision
import torchvision.transforms as transforms
import torch.utils.data as data
from PIL import Image
import os
import os.path
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union
from pathlib import Path
from collections import OrderedDict

IMG_EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp")

class Threshold_noise:
    """Remove intensities under given threshold"""

    def __init__(self, threshold=100.):
        self.threshold = threshold

    def __call__(self, x):
        return torch.where((x <= self.threshold), 0.,x)

class Log_normalisation:
    """Log normalisation of intensities"""

    def __init__(self, eps=1e-5):
        self.epsilon = eps

    def __call__(self, x):
        return torch.log(x+1+self.epsilon)/torch.log(torch.max(x)+1+self.epsilon)

class Random_shift_rt():
    pass


def load_data(base_dir, batch_size, shuffle=True, noise_threshold=0):
    train_transform = transforms.Compose(
        [transforms.Grayscale(num_output_channels=1),
         transforms.ToTensor(),
         transforms.Resize((224,224)),
         Threshold_noise(noise_threshold),
         Log_normalisation(),
         transforms.Normalize(0.5, 0.5)])
    print('Default train transform')

    val_transform = transforms.Compose(
        [transforms.Grayscale(num_output_channels=1),
         transforms.ToTensor(),
         transforms.Resize((224,224)),
         Threshold_noise(noise_threshold),
         Log_normalisation(),
         transforms.Normalize(0.5, 0.5)])
    print('Default val transform')
    train_dataset = torchvision.datasets.ImageFolder(root=base_dir, transform=train_transform)
    val_dataset = torchvision.datasets.ImageFolder(root=base_dir, transform=val_transform)
    generator1 = torch.Generator().manual_seed(42)
    indices = torch.randperm(len(train_dataset),generator=generator1)
    val_size = len(train_dataset) // 5
    train_dataset = torch.utils.data.Subset(train_dataset, indices[:-val_size])
    val_dataset = torch.utils.data.Subset(val_dataset, indices[-val_size:])

    data_loader_train = data.DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
    )

    data_loader_test = data.DataLoader(
        dataset=val_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
    )

    return data_loader_train, data_loader_test




def default_loader(path):
    return Image.open(path).convert('RGB')

def remove_aer_ana(l):
    l = map(lambda x : x.split('_')[0],l)
    return list(OrderedDict.fromkeys(l))

def make_dataset_custom(
    directory: Union[str, Path],
    class_to_idx: Optional[Dict[str, int]] = None,
    extensions: Optional[Union[str, Tuple[str, ...]]] = IMG_EXTENSIONS,
    is_valid_file: Optional[Callable[[str], bool]] = None,
    allow_empty: bool = False,
) -> List[Tuple[str, str, int]]:
    """Generates a list of samples of a form (path_to_sample, class).

    See :class:`DatasetFolder` for details.

    Note: The class_to_idx parameter is here optional and will use the logic of the ``find_classes`` function
    by default.
    """
    directory = os.path.expanduser(directory)

    if class_to_idx is None:
        _, class_to_idx = torchvision.datasets.folder.find_classes(directory)
    elif not class_to_idx:
        raise ValueError("'class_to_index' must have at least one entry to collect any samples.")

    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")

    if extensions is not None:

        def is_valid_file(x: str) -> bool:
            return torchvision.datasets.folder.has_file_allowed_extension(x, extensions)  # type: ignore[arg-type]

    is_valid_file = cast(Callable[[str], bool], is_valid_file)

    instances = []
    available_classes = set()
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            fnames_base = remove_aer_ana(fnames)
            for fname in sorted(fnames_base):
                fname_ana = fname+'_ANA.png'
                fname_aer = fname + '_AER.png'
                path_ana = os.path.join(root, fname_ana)
                path_aer = os.path.join(root, fname_aer)
                if is_valid_file(path_ana) and is_valid_file(path_aer) and os.path.isfile(path_ana) and os.path.isfile(path_aer):
                    item = path_aer, path_ana, class_index
                    instances.append(item)

                    if target_class not in available_classes:
                        available_classes.add(target_class)

    empty_classes = set(class_to_idx.keys()) - available_classes
    if empty_classes and not allow_empty:
        msg = f"Found no valid file for the classes {', '.join(sorted(empty_classes))}. "
        if extensions is not None:
            msg += f"Supported extensions are: {extensions if isinstance(extensions, str) else ', '.join(extensions)}"
        raise FileNotFoundError(msg)

    return instances


class ImageFolderDuo(data.Dataset):
    def __init__(self, root, transform=None, target_transform=None,
                 flist_reader=make_dataset_custom, loader=default_loader):
        self.root = root
        self.imlist = flist_reader(root)
        self.transform = transform
        self.target_transform = target_transform
        self.loader = loader
        self.classes = torchvision.datasets.folder.find_classes(root)[0]

    def __getitem__(self, index):
        impathAER, impathANA, target = self.imlist[index]
        imgAER = self.loader(impathAER)
        imgANA = self.loader(impathANA)
        if self.transform is not None:
            imgAER = self.transform(imgAER)
            imgANA = self.transform(imgANA)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return imgAER, imgANA, target

    def __len__(self):
        return len(self.imlist)

def load_data_duo(base_dir, batch_size, shuffle=True, noise_threshold=0):
    train_transform = transforms.Compose(
        [transforms.Grayscale(num_output_channels=1),
         transforms.ToTensor(),
         transforms.Resize((224, 224)),
         Threshold_noise(noise_threshold),
         Log_normalisation(),
         transforms.Normalize(0.5, 0.5)])
    print('Default train transform')

    val_transform = transforms.Compose(
        [transforms.Grayscale(num_output_channels=1),
         transforms.ToTensor(),
         transforms.Resize((224, 224)),
         Threshold_noise(noise_threshold),
         Log_normalisation(),
         transforms.Normalize(0.5, 0.5)])
    print('Default val transform')
    train_dataset = ImageFolderDuo(root=base_dir, transform=train_transform)
    val_dataset = ImageFolderDuo(root=base_dir, transform=val_transform)
    generator1 = torch.Generator().manual_seed(42)
    indices = torch.randperm(len(train_dataset), generator=generator1)
    val_size = len(train_dataset) // 5
    train_dataset = torch.utils.data.Subset(train_dataset, indices[:-val_size])
    val_dataset = torch.utils.data.Subset(val_dataset, indices[-val_size:])

    data_loader_train = data.DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
    )

    data_loader_test = data.DataLoader(
        dataset=val_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
    )

    return data_loader_train, data_loader_test


